#include <iostream>
#include <fstream>
#include "test/Main.hpp"
#include "test/Black.hpp"
#include "test/HullWhite.hpp"
#include "test/Data.hpp"
#include "Exam/test_Exam.hpp"
#include "Exam/Exam.hpp"

using namespace std;
using namespace prb;
using namespace cfl;
using namespace cfl::test;
using namespace cfl::test::Black;
using namespace cfl::test::HullWhite;

namespace test_prb
{
  static const char * c_pDataFile = "../Exam/Output/ExamData.txt";
  static const char * c_pDataProject = "TEST OF DATA CURVES";
  static const char * c_pBlackStandardFile = 
    "../Exam/Output/ExamBlackStandard.txt";
  static const char * c_pBlackStandardProject = 
    "TEST OF STANDARD OPTIONS IN BLACK MODEL FOR SINGLE ASSET";
  static const char * c_pHullWhiteStandardFile = 
    "../Exam/Output/ExamHullWhiteStandard.txt";
  static const char * c_pHullWhiteStandardProject = 
    "TEST OF STANDARD OPTIONS IN HULL AND WHITE  MODEL FOR INTEREST RATES";
  // static const char * c_pBlackPathDependentFile = 
  //   "../Exam/Output/ExamBlackPathDependent.txt";
  // static const char * c_pBlackPathDependentProject = 
  //   "TEST OF PATH DEPENDENT OPTIONS IN BLACK MODEL FOR SINGLE ASSET";
  static const char * c_pHullWhitePathDependentFile = 
    "../Exam/Output/ExamHullWhitePathDependent.txt";
  static const char * c_pHullWhitePathDependentProject = 
    "TEST OF PATH DEPENDENT OPTIONS IN HULL AND WHITE  MODEL FOR INTEREST RATES";

  void data();
  void standard(const cfl::Black::Data & rData);
  void standard(const cfl::HullWhite::Data & rData);
  //void pathDependent(const cfl::Black::Data & rData);
  void pathDependent(const  cfl::HullWhite::Data & rData);
}

int main()
{
  using namespace cfl::test;
  using namespace std;
  try{
    cfl::test::Data::project(test_prb::data, test_prb::c_pDataProject, 
			     test_prb::c_pDataFile);
    cfl::test::pause();
    cfl::test::Black::project(test_prb::standard, 
			      test_prb::c_pBlackStandardProject, 
			      test_prb::c_pBlackStandardFile);

    cfl::test::pause();
    cfl::test::HullWhite::project(test_prb::standard, 
				  test_prb::c_pHullWhiteStandardProject, 
				  test_prb::c_pHullWhiteStandardFile);
    cfl::test::pause();
    // cfl::test::Black::project(test_prb::pathDependent, 
    // 			      test_prb::c_pBlackPathDependentProject, 
    // 			      test_prb::c_pBlackPathDependentFile);
    // cfl::test::pause();
    cfl::test::HullWhite::project(test_prb::pathDependent, 
    				  test_prb::c_pHullWhitePathDependentProject, 
    				  test_prb::c_pHullWhitePathDependentFile);
    cfl::test::pause();
  }
  catch(const std::exception & rError) {
    std::cerr<< rError.what() << endl;
    cfl::test::pause();
  }
}

void test_prb::data()
{
  prb::test::forwardFXFit();
}

void test_prb::standard(const cfl::Black::Data & rData)
{
  double dQuality = cfl::test::Black::c_dQuality;
  cout << "quality = " << dQuality << endl;
  double dSpot = rData.forward()(rData.initialTime());
  double dInterval = cfl::test::Black::c_dInterval;
  unsigned iPoints = cfl::test::Black::c_iPoints;

  AssetModel uModel = cfl::Black::model(rData, dInterval, dQuality);
  using namespace prb::test;
  
  report(prb::test::downAndRebateOrUpAndOut, dSpot, dInterval, iPoints, uModel);
}

void test_prb::standard(const cfl::HullWhite::Data & rData)
{
  double dQuality = cfl::test::HullWhite::c_dQuality;
  cout << "quality = " << dQuality << endl;
  const double c_dDay = 1./365.;
  double dRate = - 
    std::log(rData.discount()(rData.initialTime()+c_dDay))/c_dDay;
  double dInterval = cfl::test::HullWhite::c_dInterval;
  unsigned iPoints = cfl::test::HullWhite::c_iPoints;

  InterestRateModel uModel = cfl::HullWhite::model(rData, dInterval, dQuality);

  report(prb::test::capOnSwapRateArrears, dRate, dInterval, iPoints, uModel); 
}

// void test_prb::pathDependent(const cfl::Black::Data & rData)
// {

//   double dQuality = cfl::test::Black::c_dQuality;
//   cout << "quality = " << dQuality << endl;
//   double dPathDependQuality =
//     5*cfl::test::Black::c_dPathDependQuality;
//   cout << "quality for extra state = " << dPathDependQuality << endl;
//   double dSpot = rData.forward()(rData.initialTime());
//   double dInterval = cfl::test::Black::c_dInterval;
//   unsigned iPoints = cfl::test::Black::c_iPoints;

//   AssetModel uModel = 
//     cfl::Black::model(rData, dInterval, dQuality, dPathDependQuality);

//   report(prb::test::ratchet, dSpot, dInterval, iPoints, uModel);
// }

void test_prb::pathDependent(const cfl::HullWhite::Data & rData)
{
  double dQuality = cfl::test::HullWhite::c_dQuality;
  double dPathDependQuality =
    cfl::test::HullWhite::c_dPathDependQuality*10000;
  
  cout << "quality = " << dQuality << endl;
  cout << "quality for additional state processes = " <<
    dPathDependQuality << endl;
  const double c_dDay = 1./365.;
  double dRate = 
    - std::log(rData.discount()(rData.initialTime()+c_dDay))/c_dDay;
  double dInterval = cfl::test::HullWhite::c_dInterval;
  unsigned iPoints = cfl::test::HullWhite::c_iPoints;

  InterestRateModel uModel =
    cfl::HullWhite::model(rData, dInterval, dQuality, dPathDependQuality);

  report(prb::test::deferredStartAverageRateSwap, dRate, dInterval, iPoints, uModel);
}

// DATA

void prb::test::forwardFXFit() 
{
  cfl::test::print("FORWARD PRICES FOR EXCHANGE RATES BY LEAST SQUARE FIT");
  cout << endl;
  double dSpotFX = 100;
  double dInitialTime = 1.;
  unsigned iSize = 12;
  double dPeriod = 0.25;
  std::vector<double> uDomestic(iSize);
  std::vector<double> uForeign(iSize);
  std::vector<double> uMaturity(iSize);
  double dDom = 0.04;
  double dFor = 0.1;
  
  for (unsigned int i=0;i<iSize;i++)
    {
      uMaturity[i]=dInitialTime+(i+1)*dPeriod;
      uDomestic[i] =1./(1. + dDom*(i+1)*dPeriod);
      uForeign[i] = 1./(1. + dFor*(i+1)*dPeriod);
    }
  
  cout << "initial time = " << dInitialTime << endl;
  cout << "spot FX rate = " << dSpotFX << endl;
  cout << "domestic discount factors = " << endl;
  cfl::test::print(uDomestic.begin(),uDomestic.end());
  cout << "foreign discount factors = " << endl;
  cfl::test::print(uForeign.begin(),uForeign.end());
  cout << "maturities = " << endl;
  cfl::test::print(uMaturity.begin(),uMaturity.end());
  Function uForwardFX = prb::forwardFXFit(dSpotFX, uMaturity, uDomestic, 
					  uForeign, dInitialTime);
  double dInterval = 0.5;
  cout << "output forward FX rates: " << endl;
  cfl::test::Data::print(uForwardFX, dInitialTime, dInterval); 	
}

// ASSET STANDARD

MultiFunction 
prb::test::downAndRebateOrUpAndOut(AssetModel & rModel) 
{
  cfl::test::print("DOWN-AND-REBATE-OR-UP-AND-OUT OPTION IN ASSET MODEL");
  cout << endl;
  double dLowerBarrier = cfl::test::Black::c_dLowerBarrier;
  double dUpperBarrier = cfl::test::Black::c_dUpperBarrier;
  double dNotional = cfl::test::Black::c_dNotional;
  const std::vector<double> uLowerBarrierTimes = cfl::test::Black::c_uBarrierTimes;
  const std::vector<double> uUpperBarrierTimes = cfl::test::Black::c_uExerciseTimes;
  cout << "lower barrier = " << dLowerBarrier << endl;
  cout << "upper barrier = " << dUpperBarrier << endl;
  cout << "notional = " << dNotional << endl;
  cout << "lower barrier times = "<< endl;
  cfl::test::print(uLowerBarrierTimes.begin(), uLowerBarrierTimes.end());
  cout << "upper barrier times = " << endl;
  cfl::test::print(uUpperBarrierTimes.begin(), uUpperBarrierTimes.end());
  return prb::downAndRebateOrUpAndOut(dNotional, dLowerBarrier, dUpperBarrier, 
				      uLowerBarrierTimes, uUpperBarrierTimes, 
				      rModel);
}

// INTEREST RATE STANDARD

cfl::MultiFunction 
prb::test::capOnSwapRateArrears(InterestRateModel & rModel)
{
  cfl::test::print("CAP ON SWAP RATE SET IN ARREARS IN INTEREST RATE MODEL");
  cout << endl;
  cfl::Data::CashFlow uCapParameters = cfl::test::HullWhite::swapParameters();
  uCapParameters.rate = cfl::test::HullWhite::c_dCapRate;
  double dSwapPeriod = uCapParameters.period*2;
  unsigned iSwapPayments = uCapParameters.numberOfPayments*2;
  cout << "cap parameters:" << endl;
  cfl::test::print(uCapParameters);
  cout << "swap period = " << dSwapPeriod << endl;
  cout << "number of payments in swap = " << iSwapPayments << endl;
  return prb::capOnSwapRateArrears(uCapParameters,dSwapPeriod, iSwapPayments, rModel);  
}

// ASSET PATH DEPENDENT

// INTEREST RATE PATH DEPENDENT

cfl::MultiFunction prb::test::
deferredStartAverageRateSwap(cfl::InterestRateModel & rModel, bool bPayFloat)
{
  cfl::test::print("DEFERRED START AVERAGE RATE SWAP IN INTEREST RATE MODEL");
  cout << endl;
  cfl::Data::Swap uSwapParameters = cfl::test::HullWhite::swapParameters();
  uSwapParameters.payFloat = bPayFloat;
  unsigned iResetTimes = 6;
  double dPeriod = uSwapParameters.period/ 3.33;
  std::vector<double> uResetTimes(iResetTimes,rModel.initialTime()+dPeriod);
  std::transform(uResetTimes.begin(), uResetTimes.end()-1, uResetTimes.begin()+1, 
		 [dPeriod](double dX){ return dPeriod+dX; });
  double dStartTime = uResetTimes.back() + dPeriod/2.2;
  cout << "reset times: " << endl;
  cfl::test::print(uResetTimes.begin(), uResetTimes.end());
  cout << "start time = " << dStartTime <<  endl;
  cout << "swap parameters:" << endl;
  cfl::test::print(uSwapParameters);
  return prb::deferredStartAverageRateSwap(uResetTimes, dStartTime, 
					   uSwapParameters, rModel);
}
